@@ -1,4 +1,4 @@
-// Copyright 2018-2020 Cargill Incorporated
+// Copyright 2018-2021 Cargill Incorporated
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -21,7 +21,8 @@ use std::time::Duration;
 use protobuf::Message;
 use splinter::consensus::{
     error::{ConsensusSendError, ProposalManagerError},
-    two_phase::TwoPhaseEngine,
+    two_phase::v1::TwoPhaseEngine as TwoPhaseEngineV1,
+    two_phase::v2::TwoPhaseEngine as TwoPhaseEngineV2,
     ConsensusEngine, ConsensusMessage, ConsensusNetworkSender, PeerId, Proposal, ProposalId,
     ProposalManager, ProposalUpdate, StartupState,
 };
@@ -32,6 +33,7 @@ use crate::protos::scabbard::{ProposedBatch, ScabbardMessage, ScabbardMessage_Ty
 use super::error::{ScabbardConsensusManagerError, ScabbardError};
 use super::shared::ScabbardShared;
 use super::state::ScabbardState;
+use super::ScabbardVersion;
 
 /// Component used by the service to manage and interact with consenus
 pub struct ScabbardConsensusManager {
@@ -45,6 +47,7 @@ impl ScabbardConsensusManager {
     /// consensus, and start consensus in a separate thread.
     pub fn new(
         service_id: String,
+        version: ScabbardVersion,
         shared: Arc<Mutex<ScabbardShared>>,
         state: Arc<Mutex<ScabbardState>>,
         // The coordinator timeout for the two-phase commit consensus engine
@@ -63,6 +66,7 @@ impl ScabbardConsensusManager {
 
         let proposal_manager = ScabbardProposalManager::new(
             service_id.clone(),
+            version,
             proposal_update_tx.clone(),
             shared.clone(),
             state,
@@ -77,16 +81,30 @@ impl ScabbardConsensusManager {
 
         let thread_handle = Builder::new()
             .name(format!("consensus-{}", service_id))
-            .spawn(move || {
-                let mut two_phase_engine = TwoPhaseEngine::new(coordinator_timeout);
-                if let Err(err) = two_phase_engine.run(
-                    consensus_msg_rx,
-                    proposal_update_rx,
-                    Box::new(consensus_network_sender),
-                    Box::new(proposal_manager),
-                    startup_state,
-                ) {
-                    error!("two phase consensus exited with an error: {}", err)
+            .spawn(move || match version {
+                ScabbardVersion::V1 => {
+                    let mut two_phase_engine = TwoPhaseEngineV1::new(coordinator_timeout);
+                    if let Err(err) = two_phase_engine.run(
+                        consensus_msg_rx,
+                        proposal_update_rx,
+                        Box::new(consensus_network_sender),
+                        Box::new(proposal_manager),
+                        startup_state,
+                    ) {
+                        error!("two phase consensus exited with an error: {}", err)
+                    }
+                }
+                ScabbardVersion::V2 => {
+                    let mut two_phase_engine = TwoPhaseEngineV2::new(coordinator_timeout);
+                    if let Err(err) = two_phase_engine.run(
+                        consensus_msg_rx,
+                        proposal_update_rx,
+                        Box::new(consensus_network_sender),
+                        Box::new(proposal_manager),
+                        startup_state,
+                    ) {
+                        error!("two phase consensus exited with an error: {}", err)
+                    }
                 }
             })
             .map_err(|err| ScabbardConsensusManagerError(Box::new(err)))?;
@@ -132,6 +150,7 @@ impl ScabbardConsensusManager {
 
 pub struct ScabbardProposalManager {
     service_id: String,
+    version: ScabbardVersion,
     proposal_update_sender: Sender<ProposalUpdate>,
     shared: Arc<Mutex<ScabbardShared>>,
     state: Arc<Mutex<ScabbardState>>,
@@ -140,12 +159,14 @@ pub struct ScabbardProposalManager {
 impl ScabbardProposalManager {
     pub fn new(
         service_id: String,
+        version: ScabbardVersion,
         proposal_update_sender: Sender<ProposalUpdate>,
         shared: Arc<Mutex<ScabbardShared>>,
         state: Arc<Mutex<ScabbardState>>,
     ) -> Self {
         ScabbardProposalManager {
             service_id,
+            version,
             proposal_update_sender,
             shared,
             state,
@@ -166,7 +187,10 @@ impl ProposalManager for ScabbardProposalManager {
             .lock()
             .map_err(|_| ProposalManagerError::Internal(Box::new(ScabbardError::LockPoisoned)))?;
 
-        if let Some(batch) = shared.pop_batch_from_queue() {
+        if let Some(batch) = shared
+            .pop_batch_from_queue()
+            .map_err(|err| ProposalManagerError::Internal(Box::new(err)))?
+        {
             let expected_hash = self
                 .state
                 .lock()
@@ -176,14 +200,18 @@ impl ProposalManager for ScabbardProposalManager {
 
             // Intentionally leaving out the previous_id and proposal_height fields, since this
             // service and two phase consensus don't use them. This means the proposal ID can just
-            // be the summary.
+            // be the summary (in v1) or batch ID (in v2).
+            let id = match self.version {
+                ScabbardVersion::V1 => expected_hash.as_bytes().into(),
+                ScabbardVersion::V2 => batch.batch().header_signature().as_bytes().into(),
+            };
             let proposal = Proposal {
-                id: expected_hash.as_bytes().into(),
+                id,
                 summary: expected_hash.as_bytes().into(),
                 ..Default::default()
             };
 
-            shared.add_proposed_batch(proposal.id.clone(), batch.clone());
+            shared.add_open_proposal(proposal.clone(), batch.clone());
 
             // Send the proposal to the other services
             let mut proposed_batch = ProposedBatch::new();
@@ -228,11 +256,11 @@ impl ProposalManager for ScabbardProposalManager {
     }
 
     fn check_proposal(&self, id: &ProposalId) -> Result<(), ProposalManagerError> {
-        let batch = self
+        let (proposal, batch) = self
             .shared
             .lock()
             .map_err(|_| ProposalManagerError::Internal(Box::new(ScabbardError::LockPoisoned)))?
-            .get_proposed_batch(id)
+            .get_open_proposal(id)
             .ok_or_else(|| ProposalManagerError::UnknownProposal(id.clone()))?
             .clone();
 
@@ -243,7 +271,7 @@ impl ProposalManager for ScabbardProposalManager {
             .prepare_change(batch)
             .map_err(|err| ProposalManagerError::Internal(Box::new(err)))?;
 
-        if hash.as_bytes() != id.as_ref() {
+        if hash.as_bytes() != proposal.summary {
             warn!("Hash mismatch: expected {} but was {}", id, hash);
 
             self.proposal_update_sender
@@ -268,9 +296,7 @@ impl ProposalManager for ScabbardProposalManager {
             .lock()
             .map_err(|_| ProposalManagerError::Internal(Box::new(ScabbardError::LockPoisoned)))?;
 
-        shared
-            .remove_proposed_batch(id)
-            .ok_or_else(|| ProposalManagerError::UnknownProposal(id.clone()))?;
+        shared.remove_open_proposal(id);
 
         self.state
             .lock()
@@ -292,9 +318,7 @@ impl ProposalManager for ScabbardProposalManager {
             .lock()
             .map_err(|_| ProposalManagerError::Internal(Box::new(ScabbardError::LockPoisoned)))?;
 
-        shared
-            .remove_proposed_batch(id)
-            .ok_or_else(|| ProposalManagerError::UnknownProposal(id.clone()))?;
+        shared.remove_open_proposal(id);
 
         self.state
             .lock()
@@ -380,10 +404,8 @@ mod tests {
 
     use std::collections::{HashSet, VecDeque};
 
-    use splinter::{
-        service::{ServiceMessageContext, ServiceNetworkSender, ServiceSendError},
-        signing::hash::HashVerifier,
-    };
+    use cylinder::{secp256k1::Secp256k1Context, VerifierFactory};
+    use splinter::service::{ServiceMessageContext, ServiceNetworkSender, ServiceSendError};
 
     /// Tests that the network sender properly creates messages and sends them using the
     /// `ServiceNetworkSender`.
@@ -391,20 +413,23 @@ mod tests {
     fn network_sender() {
         let service_sender = MockServiceNetworkSender::new();
         let mut peer_services = HashSet::new();
-        peer_services.insert("1".to_string());
-        peer_services.insert("2".to_string());
+        peer_services.insert("svc1".to_string());
+        peer_services.insert("svc2".to_string());
 
         let shared = Arc::new(Mutex::new(ScabbardShared::new(
             VecDeque::new(),
             Some(Box::new(service_sender.clone())),
             peer_services.clone(),
-            Box::new(HashVerifier),
+            "svc0".to_string(),
+            Secp256k1Context::new().new_verifier(),
+            #[cfg(feature = "back-pressure")]
+            ScabbardVersion::V2,
         )));
-        let consensus_sender = ScabbardConsensusNetworkSender::new("0".into(), shared);
+        let consensus_sender = ScabbardConsensusNetworkSender::new("svc0".into(), shared);
 
         // Test send_to
         consensus_sender
-            .send_to(&"1".as_bytes().into(), vec![0])
+            .send_to(&"svc1".as_bytes().into(), vec![0])
             .expect("failed to send");
 
         let (recipient, message) = service_sender
@@ -414,7 +439,7 @@ mod tests {
             .get(0)
             .expect("1st message not sent")
             .clone();
-        assert_eq!(recipient, "1".to_string());
+        assert_eq!(recipient, "svc1".to_string());
 
         let scabbard_message: ScabbardMessage =
             Message::parse_from_bytes(&message).expect("failed to parse 1st scabbard message");
@@ -427,7 +452,7 @@ mod tests {
             ConsensusMessage::try_from(scabbard_message.get_consensus_message())
                 .expect("failed to parse 1st consensus message");
         assert_eq!(consensus_message.message, vec![0]);
-        assert_eq!(consensus_message.origin_id, "0".as_bytes().into());
+        assert_eq!(consensus_message.origin_id, "svc0".as_bytes().into());
 
         // Test broadcast
         consensus_sender.broadcast(vec![1]).expect("failed to send");
@@ -453,7 +478,7 @@ mod tests {
             ConsensusMessage::try_from(scabbard_message.get_consensus_message())
                 .expect("failed to parse 2nd consensus message");
         assert_eq!(consensus_message.message, vec![1]);
-        assert_eq!(consensus_message.origin_id, "0".as_bytes().into());
+        assert_eq!(consensus_message.origin_id, "svc0".as_bytes().into());
 
         // Second broadcast message
         let (recipient, message) = service_sender
@@ -476,7 +501,7 @@ mod tests {
             ConsensusMessage::try_from(scabbard_message.get_consensus_message())
                 .expect("failed to parse 3rd consensus message");
         assert_eq!(consensus_message.message, vec![1]);
-        assert_eq!(consensus_message.origin_id, "0".as_bytes().into());
+        assert_eq!(consensus_message.origin_id, "svc0".as_bytes().into());
     }
 
     #[derive(Clone, Debug)]
