@@ -1,4 +1,4 @@
-// Copyright 2018-2020 Cargill Incorporated
+// Copyright 2018-2021 Cargill Incorporated
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -31,18 +31,20 @@ use std::path::{Path, PathBuf};
 use std::sync::{Arc, Mutex};
 use std::time::Duration;
 
+use cylinder::Verifier as SignatureVerifier;
 use openssl::hash::{hash, MessageDigest};
 use protobuf::Message;
 use splinter::{
     consensus::{Proposal, ProposalUpdate},
-    orchestrator::OrchestratableService,
     service::{
         Service, ServiceDestroyError, ServiceError, ServiceMessageContext, ServiceNetworkRegistry,
         ServiceStartError, ServiceStopError,
     },
-    signing::SignatureVerifier,
 };
-use transact::{protocol::batch::BatchPair, protos::FromBytes};
+use transact::{
+    protocol::batch::BatchPair,
+    protos::{FromBytes, IntoBytes},
+};
 
 use super::hex::to_hex;
 use super::protos::scabbard::{ScabbardMessage, ScabbardMessage_Type};
@@ -52,6 +54,8 @@ use error::ScabbardError;
 #[cfg(feature = "service-arg-validation")]
 pub use factory::ScabbardArgValidator;
 pub use factory::ScabbardFactory;
+#[cfg(feature = "factory-builder")]
+pub use factory::ScabbardFactoryBuilder;
 use shared::ScabbardShared;
 pub use state::{
     BatchInfo, BatchInfoIter, BatchStatus, Events, StateChange, StateChangeEvent, StateIter,
@@ -62,11 +66,32 @@ const SERVICE_TYPE: &str = "scabbard";
 
 const DEFAULT_COORDINATOR_TIMEOUT: u64 = 30; // 30 seconds
 
+/// Specifies the version of scabbard to use.
+#[derive(Clone, Copy, PartialEq)]
+pub enum ScabbardVersion {
+    V1,
+    V2,
+}
+
+impl TryFrom<Option<&str>> for ScabbardVersion {
+    type Error = String;
+
+    fn try_from(str_opt: Option<&str>) -> Result<Self, Self::Error> {
+        match str_opt {
+            Some("1") => Ok(Self::V1),
+            Some("2") => Ok(Self::V2),
+            Some(v) => Err(format!("Unsupported scabbard version: {}", v)),
+            None => Ok(Self::V1),
+        }
+    }
+}
+
 /// A service for running Sawtooth Sabre smart contracts with two-phase commit consensus.
 #[derive(Clone)]
 pub struct Scabbard {
     circuit_id: String,
     service_id: String,
+    version: ScabbardVersion,
     shared: Arc<Mutex<ScabbardShared>>,
     state: Arc<Mutex<ScabbardState>>,
     /// The coordinator timeout for the two-phase commit consensus engine
@@ -80,6 +105,8 @@ impl Scabbard {
     pub fn new(
         service_id: String,
         circuit_id: &str,
+        // The protocol version for scabbard
+        version: ScabbardVersion,
         // List of other scabbard services on the same circuit that this service shares state with
         peer_services: HashSet<String>,
         // The directory in which to create sabre's LMDB database
@@ -97,14 +124,22 @@ impl Scabbard {
         // default value will be used (30 seconds).
         coordinator_timeout: Option<Duration>,
     ) -> Result<Self, ScabbardError> {
-        let shared = ScabbardShared::new(VecDeque::new(), None, peer_services, signature_verifier);
+        let shared = ScabbardShared::new(
+            VecDeque::new(),
+            None,
+            peer_services,
+            service_id.clone(),
+            signature_verifier,
+            #[cfg(feature = "back-pressure")]
+            version,
+        );
 
         let (state_db_path, receipt_db_path) =
             compute_db_paths(&service_id, circuit_id, state_db_dir, receipt_db_dir)?;
         let state = ScabbardState::new(
-            state_db_path.as_path(),
+            &state_db_path,
             state_db_size,
-            receipt_db_path.as_path(),
+            &receipt_db_path,
             receipt_db_size,
             admin_keys,
         )
@@ -116,6 +151,7 @@ impl Scabbard {
         Ok(Scabbard {
             circuit_id: circuit_id.to_string(),
             service_id,
+            version,
             shared: Arc::new(Mutex::new(shared)),
             state: Arc::new(Mutex::new(state)),
             coordinator_timeout,
@@ -154,6 +190,20 @@ impl Scabbard {
             .to_string())
     }
 
+    /// Get whether the service is currently accepting batches
+    #[cfg(feature = "back-pressure")]
+    pub fn accepting_batches(&self) -> Result<bool, ScabbardError> {
+        let shared = self
+            .shared
+            .lock()
+            .map_err(|_| ScabbardError::LockPoisoned)?;
+
+        match self.version {
+            ScabbardVersion::V1 => Ok(true),
+            ScabbardVersion::V2 => Ok(shared.accepting_batches()),
+        }
+    }
+
     pub fn add_batches(&self, batches: Vec<BatchPair>) -> Result<Option<String>, ScabbardError> {
         let mut shared = self
             .shared
@@ -171,10 +221,35 @@ impl Scabbard {
                     .lock()
                     .map_err(|_| ScabbardError::LockPoisoned)?
                     .batch_history()
-                    .add_batch(batch.batch().header_signature());
+                    .add_batch(&batch.batch().header_signature());
 
                 link.push_str(&format!("{},", batch.batch().header_signature()));
-                shared.add_batch_to_queue(batch);
+
+                match self.version {
+                    ScabbardVersion::V1 => shared.add_batch_to_queue(batch)?,
+                    ScabbardVersion::V2 => {
+                        if shared.is_coordinator() {
+                            shared.add_batch_to_queue(batch)?;
+                        } else {
+                            let batch_bytes = batch
+                                .into_bytes()
+                                .map_err(|err| ScabbardError::Internal(Box::new(err)))?;
+
+                            let mut msg = ScabbardMessage::new();
+                            msg.set_message_type(ScabbardMessage_Type::NEW_BATCH);
+                            msg.set_new_batch(batch_bytes);
+                            let msg_bytes = msg
+                                .write_to_bytes()
+                                .map_err(|err| ScabbardError::Internal(Box::new(err)))?;
+
+                            shared
+                                .network_sender()
+                                .ok_or(ScabbardError::NotConnected)?
+                                .send(shared.coordinator_service_id(), msg_bytes.as_slice())
+                                .map_err(|err| ScabbardError::Internal(Box::new(err)))?;
+                        }
+                    }
+                }
             }
 
             // Remove trailing comma
@@ -258,11 +333,14 @@ impl Service for Scabbard {
         consensus.replace(
             ScabbardConsensusManager::new(
                 self.service_id().into(),
+                self.version,
                 self.shared.clone(),
                 self.state.clone(),
                 self.coordinator_timeout,
             )
-            .map_err(|err| ServiceStartError::Internal(Box::new(ScabbardError::from(err))))?,
+            .map_err(|err| {
+                ServiceStartError::Internal(format!("Unable to start consensus: {}", err))
+            })?,
         );
 
         Ok(())
@@ -312,6 +390,16 @@ impl Service for Scabbard {
         }
     }
 
+    fn purge(&mut self) -> Result<(), splinter::error::InternalError> {
+        self.state
+            .lock()
+            .map_err(|_| {
+                splinter::error::InternalError::with_message("consensus lock poisoned".into())
+            })?
+            .remove_db_files()
+            .map_err(|err| splinter::error::InternalError::from_source(Box::new(err)))
+    }
+
     fn handle_message(
         &self,
         message_bytes: &[u8],
@@ -338,7 +426,7 @@ impl Service for Scabbard {
                 self.shared
                     .lock()
                     .map_err(|_| ServiceError::PoisonedLock("shared lock poisoned".into()))?
-                    .add_proposed_batch(proposal.id.clone(), batch);
+                    .add_open_proposal(proposal.clone(), batch);
 
                 self.consensus
                     .lock()
@@ -351,7 +439,71 @@ impl Service for Scabbard {
                     ))
                     .map_err(|err| ServiceError::UnableToHandleMessage(Box::new(err)))
             }
-            ScabbardMessage_Type::UNSET => Err(ServiceError::InvalidMessageFormat(Box::new(
+            ScabbardMessage_Type::NEW_BATCH => {
+                match self.version {
+                    ScabbardVersion::V1 => {
+                        warn!("Scabbard V1 does not accept NEW_BATCH messages");
+                    }
+                    ScabbardVersion::V2 => {
+                        let mut shared = self.shared.lock().map_err(|_| {
+                            ServiceError::PoisonedLock("shared lock poisoned".into())
+                        })?;
+
+                        if shared.is_coordinator() {
+                            let batch =
+                                BatchPair::from_bytes(message.get_new_batch()).map_err(|err| {
+                                    ServiceError::UnableToHandleMessage(Box::new(err))
+                                })?;
+                            shared.add_batch_to_queue(batch).map_err(|err| {
+                                ServiceError::UnableToHandleMessage(Box::new(err))
+                            })?;
+                        } else {
+                            warn!("Ignoring new batch; this service is not the coordinator");
+                        }
+                    }
+                }
+
+                Ok(())
+            }
+            #[cfg(feature = "back-pressure")]
+            ScabbardMessage_Type::TOO_MANY_REQUESTS => {
+                match self.version {
+                    ScabbardVersion::V1 => {
+                        warn!("Scabbard V1 does not accept TOO_MANY_REQUESTS messages");
+                    }
+                    ScabbardVersion::V2 => {
+                        let mut shared = self.shared.lock().map_err(|_| {
+                            ServiceError::PoisonedLock("shared lock poisoned".into())
+                        })?;
+                        if shared.is_coordinator() {
+                            warn!("Ignoring too many requests message, not from the coordinator");
+                        } else {
+                            shared.set_accepting_batches(false);
+                        }
+                    }
+                }
+                Ok(())
+            }
+            #[cfg(feature = "back-pressure")]
+            ScabbardMessage_Type::ACCEPTING_REQUESTS => {
+                match self.version {
+                    ScabbardVersion::V1 => {
+                        warn!("Scabbard V1 does not accept ACCEPTING_REQUESTS messages");
+                    }
+                    ScabbardVersion::V2 => {
+                        let mut shared = self.shared.lock().map_err(|_| {
+                            ServiceError::PoisonedLock("shared lock poisoned".into())
+                        })?;
+                        if shared.is_coordinator() {
+                            warn!("Ignoring accepting requests message, not from the coordinator");
+                        } else {
+                            shared.set_accepting_batches(true);
+                        }
+                    }
+                }
+                Ok(())
+            }
+            _ => Err(ServiceError::InvalidMessageFormat(Box::new(
                 ScabbardError::MessageTypeUnset,
             ))),
         }
@@ -362,16 +514,6 @@ impl Service for Scabbard {
     }
 }
 
-impl OrchestratableService for Scabbard {
-    fn as_service(&self) -> &dyn Service {
-        self
-    }
-
-    fn clone_box(&self) -> Box<dyn OrchestratableService> {
-        Box::new(self.clone())
-    }
-}
-
 fn compute_db_paths(
     service_id: &str,
     circuit_id: &str,
@@ -384,8 +526,8 @@ fn compute_db_paths(
     )
     .map(|digest| to_hex(&*digest))
     .map_err(|err| ScabbardError::InitializationFailed(Box::new(err)))?;
-    let state_db_path = state_db_dir.join(format!("{}-state.lmdb", hash));
-    let receipt_db_path = receipt_db_dir.join(format!("{}-receipts.lmdb", hash));
+    let state_db_path = state_db_dir.join(format!("{}-state", hash));
+    let receipt_db_path = receipt_db_dir.join(format!("{}-receipts", hash));
     Ok((state_db_path, receipt_db_path))
 }
 
@@ -395,12 +537,10 @@ pub mod tests {
 
     use std::error::Error;
 
-    use splinter::{
-        service::{
-            ServiceConnectionError, ServiceDisconnectionError, ServiceMessageContext,
-            ServiceNetworkSender, ServiceSendError,
-        },
-        signing::hash::HashVerifier,
+    use cylinder::{secp256k1::Secp256k1Context, VerifierFactory};
+    use splinter::service::{
+        ServiceConnectionError, ServiceDisconnectionError, ServiceMessageContext,
+        ServiceNetworkSender, ServiceSendError,
     };
 
     /// Tests that a new scabbard service is properly instantiated.
@@ -409,12 +549,13 @@ pub mod tests {
         let service = Scabbard::new(
             "new_scabbard".into(),
             "test_circuit",
+            ScabbardVersion::V1,
             HashSet::new(),
             Path::new("/tmp"),
             1024 * 1024,
             Path::new("/tmp"),
             1024 * 1024,
-            Box::new(HashVerifier),
+            Secp256k1Context::new().new_verifier(),
             vec![],
             None,
         )
@@ -430,12 +571,13 @@ pub mod tests {
         let mut service = Scabbard::new(
             "thread_cleanup".into(),
             "test_circuit",
+            ScabbardVersion::V1,
             HashSet::new(),
             Path::new("/tmp"),
             1024 * 1024,
             Path::new("/tmp"),
             1024 * 1024,
-            Box::new(HashVerifier),
+            Secp256k1Context::new().new_verifier(),
             vec![],
             None,
         )
@@ -451,12 +593,13 @@ pub mod tests {
         let mut service = Scabbard::new(
             "connect_and_disconnect".into(),
             "test_circuit",
+            ScabbardVersion::V1,
             HashSet::new(),
             Path::new("/tmp"),
             1024 * 1024,
             Path::new("/tmp"),
             1024 * 1024,
-            Box::new(HashVerifier),
+            Secp256k1Context::new().new_verifier(),
             vec![],
             None,
         )
